@@ -24,7 +24,6 @@ particular solver.
 """
 
 from functools import partial
-from six import iteritems
 
 from pysmt.exceptions import (NoSolverAvailableError, SolverRedefinitionError,
                               NoLogicAvailableError,
@@ -102,7 +101,7 @@ class Factory(object):
 
 
     def get_unsat_core_solver(self, name=None, logic=None,
-                              unsat_cores_mode="all"):
+                              unsat_cores_mode="all", **options):
         SolverClass, closer_logic = \
            self._get_solver_class(solver_list=self._all_unsat_core_solvers,
                                   solver_type="Solver supporting Unsat Cores",
@@ -113,7 +112,8 @@ class Factory(object):
         return SolverClass(environment=self.environment,
                            logic=closer_logic,
                            generate_models=True,
-                           unsat_cores_mode=unsat_cores_mode)
+                           unsat_cores_mode=unsat_cores_mode,
+                           **options)
 
     def get_quantifier_eliminator(self, name=None, logic=None):
         SolverClass, closer_logic = \
@@ -271,7 +271,7 @@ class Factory(object):
         else:
             self._all_solvers = installed_solvers
 
-        for k,s in iteritems(self._all_solvers):
+        for k,s in self._all_solvers.items():
             try:
                 if s.UNSAT_CORE_SUPPORT:
                     self._all_unsat_core_solvers[k] = s
@@ -356,7 +356,7 @@ class Factory(object):
         """
         res = {}
         if logic is not None:
-            for s, v in iteritems(solver_list):
+            for s, v in solver_list.items():
                 for l in v.LOGICS:
                     if logic <= l:
                         res[s] = v
@@ -439,10 +439,11 @@ class Factory(object):
                                **options)
 
     def UnsatCoreSolver(self, name=None, logic=None,
-                        unsat_cores_mode="all"):
+                        unsat_cores_mode="all", **options):
         return self.get_unsat_core_solver(name=name,
                                           logic=logic,
-                                          unsat_cores_mode=unsat_cores_mode)
+                                          unsat_cores_mode=unsat_cores_mode,
+                                          **options)
 
     def QuantifierEliminator(self, name=None, logic=None):
         return self.get_quantifier_eliminator(name=name, logic=logic)
