@@ -19,7 +19,6 @@
 the SMTLIB and provides methods to compare and search for particular
 logics.
 """
-import six
 
 from pysmt.exceptions import UndefinedLogicError, NoLogicAvailableError
 
@@ -590,6 +589,7 @@ UFNIA = Logic(name="UFNIA",
               description=\
 """Non-linear integer arithmetic with uninterpreted sort and function
 symbols.""",
+              integer_arithmetic=True,
               integer_difference=True,
               linear=False,
               uninterpreted=True)
@@ -722,7 +722,7 @@ def convert_logic_from_string(name):
 
     This takes a logic or a string or None, and returns a logic or None.
     """
-    if name is not None and isinstance(name, six.string_types):
+    if name is not None and isinstance(name, str):
         name = get_logic_by_name(name)
     return name
 
