@@ -15,12 +15,12 @@
 #   See the License for the specific language governing permissions and
 #   limitations under the License.
 #
-import os
-from pysmt.logics import QF_LRA
-from pysmt.test.smtlib.parser_utils import execute_script_fname, SMTLIB_TEST_FILES, SMTLIB_DIR
+import pytest
 
-def test_generator():
-    for (logic, f, expected_result) in SMTLIB_TEST_FILES:
-        smtfile = os.path.join(SMTLIB_DIR, f)
-        if logic == QF_LRA:
-            yield execute_script_fname, smtfile, logic, expected_result
+from pysmt.logics import QF_LRA
+from pysmt.test.smtlib.parser_utils import execute_script_fname, smtlib_tests
+
+
+@pytest.mark.parametrize("smtfile, logic, expected_result", smtlib_tests(lambda x: x == QF_LRA))
+def test_qf_lra(smtfile, logic, expected_result):
+    execute_script_fname(smtfile, logic, expected_result)
