@@ -15,10 +15,8 @@
 #   See the License for the specific language governing permissions and
 #   limitations under the License.
 #
-from six.moves import xrange
-from six import PY2
-
 import pysmt.operators as op
+
 from pysmt.shortcuts import Symbol, FreshSymbol, And, Not, GT, Function, Plus
 from pysmt.shortcuts import Bool, TRUE, Real, LE, FALSE, Or, Equals, Implies
 from pysmt.shortcuts import Solver
@@ -155,6 +153,47 @@ class TestBasic(TestCase):
             s.solve()
             self.assertTrue(s.get_py_value(varA))
 
+    @skipIfNoSolverForLogic(QF_BOOL)
+    def test_add_assertions(self):
+        varA = Symbol("A", BOOL)
+        varB = Symbol("B", BOOL)
+        varC = Symbol("C", BOOL)
+
+        assertions = [varA, Implies(varA, varB), Implies(varB, varC)]
+        for name in get_env().factory.all_solvers(logic=QF_BOOL):
+            with Solver(name) as solver:
+                solver.add_assertions(assertions)
+                solver.solve()
+                self.assertTrue(solver.get_py_value(And(assertions)))
+
+    @skipIfNoSolverForLogic(QF_BOOL)
+    def test_incremental(self):
+        a = Symbol('a', BOOL)
+        b = Symbol('b', BOOL)
+        c = Symbol('c', BOOL)
+
+        for name in get_env().factory.all_solvers(logic=QF_BOOL):
+            with Solver(name) as solver:
+                solver.add_assertion(Or(a, b))
+                solver.add_assertion(Or(Not(b), c))
+                self.assertTrue(solver.solve())
+                try:
+                    solver.push(1)
+                except NotImplementedError:
+                    # if push not implemented, pop shouldn't be either
+                    self.assertRaises(NotImplementedError, solver.pop)
+                    continue
+
+                solver.add_assertion(And(Not(a), Not(c)))
+                self.assertFalse(solver.solve())
+                solver.pop(1)
+                self.assertTrue(solver.solve())
+                solver.add_assertion(FALSE())
+                self.assertFalse(solver.solve())
+                solver.reset_assertions()
+                solver.add_assertion(a)
+                self.assertTrue(solver.solve())
+
     @skipIfSolverNotAvailable("msat")
     def test_examples_msat(self):
         for (f, validity, satisfiability, logic) in get_example_formulae():
@@ -212,6 +251,7 @@ class TestBasic(TestCase):
             self.assertEqual(satisfiability, s, f)
 
     def do_model(self, solver_name):
+
         for (f, _, satisfiability, logic) in get_example_formulae():
             if satisfiability and not logic.theory.uninterpreted and logic.quantifier_free:
                 try:
@@ -221,24 +261,9 @@ class TestBasic(TestCase):
                         check = s.solve()
                         self.assertTrue(check)
 
-                        # Ask single values to the solver
-                        subs = {}
-                        for d in f.get_free_variables():
-                            m = s.get_value(d)
-                            subs[d] = m
-
-                        simp = f.substitute(subs).simplify()
-                        self.assertEqual(simp, TRUE(), "%s -- %s :> %s" % (f, subs, simp))
-
-                        # Ask the eager model
-                        subs = {}
                         model = s.get_model()
-                        for d in f.get_free_variables():
-                            m = model.get_value(d)
-                            subs[d] = m
+                        self.assertTrue(model.satisfies(f, s))
 
-                        simp = f.substitute(subs).simplify()
-                        self.assertEqual(simp, TRUE())
                 except NoSolverAvailableError:
                     pass
 
@@ -266,8 +291,9 @@ class TestBasic(TestCase):
     def test_tactics_z3(self):
         from z3 import Tactic, Then
 
-        my_tactic = Then(Tactic('simplify'), Tactic('propagate-values'),
-                         Tactic('elim-uncnstr'))
+        my_tactic = lambda ctx : Then(Tactic('simplify', ctx),
+                                      Tactic('propagate-values', ctx),
+                                      Tactic('elim-uncnstr', ctx))
 
         for (f, validity, satisfiability, logic) in get_example_formulae():
             if not logic.theory.linear: continue
@@ -276,7 +302,7 @@ class TestBasic(TestCase):
             if logic.theory.bit_vectors: continue
             s = Solver(name='z3')
             z3_f = s.converter.convert(f)
-            simp_z3_f = my_tactic(z3_f)
+            simp_z3_f = my_tactic(s.z3.ctx)(z3_f)
             simp_f = s.converter.back(simp_z3_f.as_expr())
             v = is_valid(simp_f)
             s = is_sat(simp_f)
@@ -323,7 +349,7 @@ class TestBasic(TestCase):
                         self.assertIsNone(f_i)
 
     def test_solving_under_assumption(self):
-        v1, v2 = [FreshSymbol() for _ in xrange(2)]
+        v1, v2 = [FreshSymbol() for _ in range(2)]
         xor = Or(And(v1, Not(v2)), And(Not(v1), v2))
 
         for name in get_env().factory.all_solvers():
@@ -521,9 +547,6 @@ class TestBasic(TestCase):
     @skipIfNoSolverForLogic(QF_LRA)
     def test_logic_as_string(self):
         self.assertEqual(convert_logic_from_string("QF_LRA"), QF_LRA)
-        if PY2:
-            self.assertEqual(convert_logic_from_string(unicode("QF_LRA")),
-                             QF_LRA)
         with self.assertRaises(UndefinedLogicError):
             convert_logic_from_string("PAPAYA")
         self.assertIsNone(convert_logic_from_string(None))
@@ -594,11 +617,12 @@ class TestBasic(TestCase):
             if logic == QF_BV:
                 solver = Solver(name="btor",
                                 solver_options={"rewrite-level":0,
-                                                "fun:dual-prop":1,
+                                                "fun-dual-prop":1,
                                                 "eliminate-slices":1})
                 solver.add_assertion(f)
                 res = solver.solve()
                 self.assertTrue(res == sat)
 
+
 if __name__ == '__main__':
     main()
